diff --git a/doc/src/sgml/xindex.sgml b/doc/src/sgml/xindex.sgml
index a90b4e2..b6fe908 100644
*** a/doc/src/sgml/xindex.sgml
--- b/doc/src/sgml/xindex.sgml
***************
*** 834,846 **** ALTER OPERATOR FAMILY integer_ops USING btree ADD
    <para>
     In a B-tree operator family, all the operators in the family must sort
     compatibly, meaning that the transitive laws hold across all the data types
!    supported by the family: <quote>if A = B and B = C, then A =
!    C</>, and <quote>if A &lt; B and B &lt; C, then A &lt; C</>.  For each
!    operator in the family there must be a support function having the same
!    two input data types as the operator.  It is recommended that a family be
!    complete, i.e., for each combination of data types, all operators are
!    included.  Each operator class should include just the non-cross-type
!    operators and support function for its data type.
    </para>
  
    <para>
--- 834,848 ----
    <para>
     In a B-tree operator family, all the operators in the family must sort
     compatibly, meaning that the transitive laws hold across all the data types
!    supported by the family: <quote>if A = B and B = C, then A = C</>,
!    and <quote>if A &lt; B and B &lt; C, then A &lt; C</>.  Subjecting operands
!    to any number of implicit casts or binary coercion casts involving only
!    types represented in the family must not change the result other than to
!    require a similar cast thereon.  For each operator in the family there must
!    be a support function having the same two input data types as the operator.
!    It is recommended that a family be complete, i.e., for each combination of
!    data types, all operators are included.  Each operator class should include
!    just the non-cross-type operators and support function for its data type.
    </para>
  
    <para>
***************
*** 851,861 **** ALTER OPERATOR FAMILY integer_ops USING btree ADD
     by the family's equality operators, even when the values are of different
     types.  This is usually difficult to accomplish when the types have
     different physical representations, but it can be done in some cases.
!    Notice that there is only one support function per data type, not one
!    per equality operator.  It is recommended that a family be complete, i.e.,
!    provide an equality operator for each combination of data types.
!    Each operator class should include just the non-cross-type equality
!    operator and the support function for its data type.
    </para>
  
    <para>
--- 853,865 ----
     by the family's equality operators, even when the values are of different
     types.  This is usually difficult to accomplish when the types have
     different physical representations, but it can be done in some cases.
!    Implicit casts and binary coercion casts among types represented in the
!    family must preserve this invariant.  Notice that there is only one support
!    function per data type, not one per equality operator.  It is recommended
!    that a family be complete, i.e., provide an equality operator for each
!    combination of data types.  Each operator class should include just the
!    non-cross-type equality operator and the support function for its data
!    type.
    </para>
  
    <para>
diff --git a/src/backend/bootstindex f3e85aa..d0a0e92 100644
*** a/src/backend/bootstrap/bootparse.y
--- b/src/backend/bootstrap/bootparse.y
***************
*** 217,222 **** Boot_CreateStmt:
--- 217,223 ----
  												   PG_CATALOG_NAMESPACE,
  												   shared_relation ? GLOBALTABLESPACE_OID : 0,
  												   $3,
+ 												   InvalidOid,
  												   tupdesc,
  												   RELKIND_RELATION,
  												   RELPERSISTENCE_PERMANENT,
***************
*** 284,289 **** Boot_DeclareIndexStmt:
--- 285,291 ----
  					DefineIndex(makeRangeVar(NULL, $6, -1),
  								$3,
  								$4,
+ 								InvalidOid,
  								$8,
  								NULL,
  								$10,
***************
*** 302,307 **** Boot_DeclareUniqueIndexStmt:
--- 304,310 ----
  					DefineIndex(makeRangeVar(NULL, $7, -1),
  								$4,
  								$5,
+ 								InvalidOid,
  								$9,
  								NULL,
  								$11,
diff --git a/src/backend/catalog/heap.c index a8c2700..45ccb04 100644
*** a/src/backend/catalog/heap.c
--- b/src/backend/catalog/heap.c
***************
*** 229,235 **** SystemAttributeByName(const char *attname, bool relhasoids)
   *		heap_create		- Create an uncataloged heap relation
   *
   *		Note API change: the caller must now always provide the OID
!  *		to use for the relation.
   *
   *		rel->rd_rel is initialized by RelationBuildLocalRelation,
   *		and is mostly zeroes at return.
--- 229,236 ----
   *		heap_create		- Create an uncataloged heap relation
   *
   *		Note API change: the caller must now always provide the OID
!  *		to use for the relation.  The relfilenode may (and, normally,
!  *		should) be left unspecified.
   *
   *		rel->rd_rel is initialized by RelationBuildLocalRelation,
   *		and is mostly zeroes at return.
***************
*** 240,245 **** heap_create(const char *relname,
--- 241,247 ----
  			Oid relnamespace,
  			Oid reltablespace,
  			Oid relid,
+ 			Oid relfilenode,
  			TupleDesc tupDesc,
  			char relkind,
  			char relpersistence,
***************
*** 297,302 **** heap_create(const char *relname,
--- 299,314 ----
  	}
  
  	/*
+ 	 * Unless otherwise requested, the physical ID (relfilenode) is initially
+ 	 * the same as the logical ID (OID).  When the caller did specify a
+ 	 * relfilenode, it already exists; do not attempt to create it.
+ 	 */
+ 	if (OidIsValid(relfilenode))
+ 		create_storage = false;
+ 	else
+ 		relfilenode = relid;
+ 
+ 	/*
  	 * Never allow a pg_class entry to explicitly specify the database's
  	 * default tablespace in reltablespace; force it to zero instead. This
  	 * ensures that if the database is cloned with a different default
***************
*** 315,320 **** heap_create(const char *relname,
--- 327,333 ----
  									 relnamespace,
  									 tupDesc,
  									 relid,
+ 									 relfilenode,
  									 reltablespace,
  									 shared_relation,
  									 mapped_relation,
***************
*** 1103,1108 **** heap_create_with_catalog(const char *relname,
--- 1116,1122 ----
  							   relnamespace,
  							   reltablespace,
  							   relid,
+ 							   InvalidOid,
  							   tupdesc,
  							   relkind,
  							   relpersistence,
diff --git a/src/backend/catalog/index 6fc234b..e32639e 100644
*** a/src/backend/catalog/index.c
--- b/src/backend/catalog/index.c
***************
*** 649,654 **** UpdateIndexRelation(Oid indexoid,
--- 649,656 ----
   * indexRelationId: normally, pass InvalidOid to let this routine
   *		generate an OID for the index.	During bootstrap this may be
   *		nonzero to specify a preselected OID.
+  * relFileNode: normally, pass InvalidOid to get new storage.  May be
+  *		nonzero to attach an existing valid build.
   * indexInfo: same info executor uses to insert into the index
   * indexColNames: column names to use for index (List of char *)
   * accessMethodObjectId: OID of index AM to use
***************
*** 674,679 **** Oid
--- 676,682 ----
  index_create(Relation heapRelation,
  			 const char *indexRelationName,
  			 Oid indexRelationId,
+ 			 Oid relFileNode,
  			 IndexInfo *indexInfo,
  			 List *indexColNames,
  			 Oid accessMethodObjectId,
***************
*** 813,818 **** index_create(Relation heapRelation,
--- 816,822 ----
  								namespaceId,
  								tableSpaceId,
  								indexRelationId,
+ 								relFileNode,
  								indexTupDesc,
  								RELKIND_INDEX,
  								relpersistence,
diff --git a/src/backend/catalog/sindex 57987be..d027361 100644
*** a/src/backend/catalog/storage.c
--- b/src/backend/catalog/storage.c
***************
*** 206,215 **** RelationDropStorage(Relation rel)
   * The relation mapper fixes this by telling us to not delete such relations
   * after all as part of its commit.
   *
   * No-op if the relation is not among those scheduled for deletion.
   */
  void
! RelationPreserveStorage(RelFileNode rnode)
  {
  	PendingRelDelete *pending;
  	PendingRelDelete *prev;
--- 206,218 ----
   * The relation mapper fixes this by telling us to not delete such relations
   * after all as part of its commit.
   *
+  * We also use this to reuse an old build of an index during ALTER TABLE, this
+  * time removing the delete-at-commit entry.
+  *
   * No-op if the relation is not among those scheduled for deletion.
   */
  void
! RelationPreserveStorage(RelFileNode rnode, bool atCommit)
  {
  	PendingRelDelete *pending;
  	PendingRelDelete *prev;
***************
*** 219,229 **** RelationPreserveStorage(RelFileNode rnode)
  	for (pending = pendingDeletes; pending != NULL; pending = next)
  	{
  		next = pending->next;
! 		if (RelFileNodeEquals(rnode, pending->relnode))
  		{
- 			/* we should only find delete-on-abort entries, else trouble */
- 			if (pending->atCommit)
- 				elog(ERROR, "cannot preserve a delete-on-commit relation");
  			/* unlink and delete list entry */
  			if (prev)
  				prev->next = next;
--- 222,230 ----
  	for (pending = pendingDeletes; pending != NULL; pending = next)
  	{
  		next = pending->next;
! 		if (RelFileNodeEquals(rnode, pending->relnode)
! 			&& pending->atCommit == atCommit)
  		{
  			/* unlink and delete list entry */
  			if (prev)
  				prev->next = next;
diff --git a/src/backend/catalog/toaindex ce082fd..a09a3ad 100644
*** a/src/backend/catalog/toasting.c
--- b/src/backend/catalog/toasting.c
***************
*** 274,280 **** create_toast_table(Relation rel, Oid toastOid, Oid toastIndexOid, Datum reloptio
  	coloptions[0] = 0;
  	coloptions[1] = 0;
  
! 	index_create(toast_rel, toast_idxname, toastIndexOid,
  				 indexInfo,
  				 list_make2("chunk_id", "chunk_seq"),
  				 BTREE_AM_OID,
--- 274,280 ----
  	coloptions[0] = 0;
  	coloptions[1] = 0;
  
! 	index_create(toast_rel, toast_idxname, toastIndexOid, InvalidOid,
  				 indexInfo,
  				 list_make2("chunk_id", "chunk_seq"),
  				 BTREE_AM_OID,
diff --git a/src/backend/commands/indindex b7c021d..ff4d1df 100644
*** a/src/backend/commands/indexcmds.c
--- b/src/backend/commands/indexcmds.c
***************
*** 73,78 **** static char *ChooseIndexNameAddition(List *colnames);
--- 73,264 ----
  
  
  /*
+  * CheckIndexCompatible
+  *		Determine whether an existing index definition is compatible with a
+  *		prospective index definition, such that the existing index storage
+  *		could become the storage of the new index, avoiding a rebuild.
+  *
+  * 'heapRelation': the relation the index would apply to.
+  * 'accessMethodName': name of the AM to use.
+  * 'attributeList': a list of IndexElem specifying columns and expressions
+  *		to index on.
+  * 'exclusionOpNames': list of names of exclusion-constraint operators,
+  *		or NIL if not an exclusion constraint.
+  *
+  * This is tailored to the needs of ALTER TABLE ALTER TYPE, which recreates
+  * any indexes that depended on a changing column from their pg_get_indexdef
+  * or pg_get_constraintdef definitions.  We omit some of the sanity checks of
+  * DefineIndex.  We assume that the old and new indexes have the same number
+  * of columns and that if one has an expression column or predicate, both do.
+  * Errors arising from the attribute list still apply.
+  *
+  * Most column type changes that can skip a table rewrite will not invalidate
+  * indexes.  For btree and hash indexes, we assume continued validity when
+  * each column of an index would have the same operator family before and
+  * after the change.  Since we do not document a contract for GIN or GiST
+  * operator families, we require an exact operator class match for them and
+  * for any other access methods.
+  *
+  * DefineIndex always verifies that each exclusion operator shares an operator
+  * family with its corresponding index operator class.  For access methods
+  * having no operator family contract, confirm that the old and new indexes
+  * use the exact same exclusion operator.  For btree and hash, there's nothing
+  * more to check.
+  *
+  * We do not yet implement a test to verify compatibility of expression
+  * columns or predicates, so assume any such index is incompatible.
+  */
+ bool
+ CheckIndexCompatible(Oid oldId,
+ 					 RangeVar *heapRelation,
+ 					 char *accessMethodName,
+ 					 List *attributeList,
+ 					 List *exclusionOpNames)
+ {
+ 	bool		isconstraint;
+ 	Oid		   *collationObjectId;
+ 	Oid		   *classObjectId;
+ 	Oid			accessMethodId;
+ 	Oid			relationId;
+ 	HeapTuple	tuple;
+ 	Form_pg_am	accessMethodForm;
+ 	bool		amcanorder;
+ 	RegProcedure amoptions;
+ 	int16	   *coloptions;
+ 	IndexInfo  *indexInfo;
+ 	int			numberOfAttributes;
+ 	int			old_natts;
+ 	bool		isnull;
+ 	bool		family_am;
+ 	bool		ret = true;
+ 	oidvector  *old_indclass;
+ 	oidvector  *old_indcollation;
+ 	int			i;
+ 	Datum		d;
+ 
+ 	/* Caller should already have the relation locked in some way. */
+ 	relationId = RangeVarGetRelid(heapRelation, false);
+ 	/*
+ 	 * We can pretend isconstraint = false unconditionally.  It only serves to
+ 	 * decide the text of an error message that should never happen for us.
+ 	 */
+ 	isconstraint = false;
+ 
+ 	numberOfAttributes = list_length(attributeList);
+ 	Assert(numberOfAttributes > 0);
+ 	Assert(numberOfAttributes <= INDEX_MAX_KEYS);
+ 
+ 	/* look up the access method */
+ 	tuple = SearchSysCache1(AMNAME, PointerGetDatum(accessMethodName));
+ 	if (!HeapTupleIsValid(tuple))
+ 		ereport(ERROR,
+ 				(errcode(ERRCODE_UNDEFINED_OBJECT),
+ 				 errmsg("access method \"%s\" does not exist",
+ 						accessMethodName)));
+ 	accessMethodId = HeapTupleGetOid(tuple);
+ 	accessMethodForm = (Form_pg_am) GETSTRUCT(tuple);
+ 	amcanorder = accessMethodForm->amcanorder;
+ 	amoptions = accessMethodForm->amoptions;
+ 	ReleaseSysCache(tuple);
+ 
+ 	indexInfo = makeNode(IndexInfo);
+ 	indexInfo->ii_Expressions = NIL;
+ 	indexInfo->ii_ExpressionsState = NIL;
+ 	indexInfo->ii_PredicateState = NIL;
+ 	indexInfo->ii_ExclusionOps = NULL;
+ 	indexInfo->ii_ExclusionProcs = NULL;
+ 	indexInfo->ii_ExclusionStrats = NULL;
+ 	collationObjectId = (Oid *) palloc(numberOfAttributes * sizeof(Oid));
+ 	classObjectId = (Oid *) palloc(numberOfAttributes * sizeof(Oid));
+ 	coloptions = (int16 *) palloc(numberOfAttributes * sizeof(int16));
+ 
+ 	ComputeIndexAttrs(indexInfo, collationObjectId, classObjectId,
+ 					  coloptions, attributeList,
+ 					  exclusionOpNames, relationId,
+ 					  accessMethodName, accessMethodId,
+ 					  amcanorder, isconstraint);
+ 
+ 
+ 	/* Get the soon-obsolete pg_index tuple. */
+ 	tuple = SearchSysCache1(INDEXRELID, ObjectIdGetDatum(oldId));
+ 	if (!HeapTupleIsValid(tuple))
+ 		elog(ERROR, "cache lookup failed for index %u", oldId);
+ 
+ 	/* We don't assess expressions or predicates; assume incompatibility. */
+ 	if (!(heap_attisnull(tuple, Anum_pg_index_indpred) &&
+ 		  heap_attisnull(tuple, Anum_pg_index_indexprs)))
+ 	{
+ 		ReleaseSysCache(tuple);
+ 		return false;
+ 	}
+ 
+ 	/*
+ 	 * If the old and new operator class of any index column differ in
+ 	 * operator family or collation, regard the old index as incompatible.
+ 	 * For access methods other than btree and hash, a family match has no
+ 	 * defined meaning; require an exact operator class match.
+ 	 */
+ 	old_natts = ((Form_pg_index) GETSTRUCT(tuple))->indnatts;
+ 	Assert(old_natts == numberOfAttributes);
+ 
+ 	d = SysCacheGetAttr(INDEXRELID, tuple, Anum_pg_index_indcollation, &isnull);
+ 	Assert(!isnull);
+ 	old_indcollation = (oidvector *) DatumGetPointer(d);
+ 
+ 	d = SysCacheGetAttr(INDEXRELID, tuple, Anum_pg_index_indclass, &isnull);
+ 	Assert(!isnull);
+ 	old_indclass = (oidvector *) DatumGetPointer(d);
+ 
+ 	family_am = accessMethodId == BTREE_AM_OID || accessMethodId == HASH_AM_OID;
+ 
+ 	for (i = 0; i < old_natts; i++)
+ 	{
+ 		Oid			old_class = old_indclass->values[i];
+ 		Oid			new_class = classObjectId[i];
+ 
+ 		if (!(old_indcollation->values[i] == collationObjectId[i]
+ 			  && (old_class == new_class
+ 				  || (family_am && (get_opclass_family(old_class)
+ 									== get_opclass_family(new_class))))))
+ 		{
+ 			ret = false;
+ 			break;
+ 		}
+ 	}
+ 
+ 	ReleaseSysCache(tuple);
+ 
+ 	/*
+ 	 * For btree and hash, exclusion operators need only fall in the same
+ 	 * operator family; ComputeIndexAttrs already verified that much.  If we
+ 	 * get this far, we know that the index operator family has not changed,
+ 	 * and we're done.  For other access methods, require exact matches for
+ 	 * all exclusion operators.
+ 	 */
+ 	if (ret && !family_am && indexInfo->ii_ExclusionOps != NULL)
+ 	{
+ 		Relation	irel;
+ 		Oid		   *old_operators, *old_procs;
+ 		uint16	   *old_strats;
+ 
+ 		/* Caller probably already holds a stronger lock. */
+ 		irel = index_open(oldId, AccessShareLock);
+ 		RelationGetExclusionInfo(irel, &old_operators, &old_procs, &old_strats);
+ 
+ 		for (i = 0; i < old_natts; i++)
+ 			if (old_operators[i] != indexInfo->ii_ExclusionOps[i])
+ 			{
+ 				ret = false;
+ 				break;
+ 			}
+ 
+ 		index_close(irel, NoLock);
+ 	}
+ 
+ 	return ret;
+ }
+ 
+ /*
   * DefineIndex
   *		Creates a new index.
   *
***************
*** 81,86 **** static char *ChooseIndexNameAddition(List *colnames);
--- 267,274 ----
   *		that a nonconflicting default name should be picked.
   * 'indexRelationId': normally InvalidOid, but during bootstrap can be
   *		nonzero to specify a preselected OID for the index.
+  * 'relFileNode': normally InvalidOid, but can be nonzero to specify existing
+  *		storage constituting a valid build of this index.
   * 'accessMethodName': name of the AM to use.
   * 'tableSpaceName': name of the tablespace to create the index in.
   *		NULL specifies using the appropriate default.
***************
*** 103,113 **** static char *ChooseIndexNameAddition(List *colnames);
   *		it will be filled later.
   * 'quiet': suppress the NOTICE chatter ordinarily provided for constraints.
   * 'concurrent': avoid blocking writers to the table while building.
   */
! void
  DefineIndex(RangeVar *heapRelation,
  			char *indexRelationName,
  			Oid indexRelationId,
  			char *accessMethodName,
  			char *tableSpaceName,
  			List *attributeList,
--- 291,304 ----
   *		it will be filled later.
   * 'quiet': suppress the NOTICE chatter ordinarily provided for constraints.
   * 'concurrent': avoid blocking writers to the table while building.
+  *
+  * Returns the OID of the created index.
   */
! Oid
  DefineIndex(RangeVar *heapRelation,
  			char *indexRelationName,
  			Oid indexRelationId,
+ 			Oid relFileNode,
  			char *accessMethodName,
  			char *tableSpaceName,
  			List *attributeList,
***************
*** 403,413 **** DefineIndex(RangeVar *heapRelation,
  	}
  
  	/*
  	 * Make the catalog entries for the index, including constraints. Then, if
  	 * not skip_build || concurrent, actually build the index.
  	 */
  	indexRelationId =
! 		index_create(rel, indexRelationName, indexRelationId,
  					 indexInfo, indexColNames,
  					 accessMethodId, tablespaceId,
  					 collationObjectId, classObjectId,
--- 594,610 ----
  	}
  
  	/*
+ 	 * A valid relFileNode implies that we already have a built form of the
+ 	 * index.  The caller should also decline any index build.
+ 	 */
+ 	Assert(!OidIsValid(relFileNode) || (skip_build && !concurrent));
+ 
+ 	/*
  	 * Make the catalog entries for the index, including constraints. Then, if
  	 * not skip_build || concurrent, actually build the index.
  	 */
  	indexRelationId =
! 		index_create(rel, indexRelationName, indexRelationId, relFileNode,
  					 indexInfo, indexColNames,
  					 accessMethodId, tablespaceId,
  					 collationObjectId, classObjectId,
***************
*** 421,427 **** DefineIndex(RangeVar *heapRelation,
  	{
  		/* Close the heap and we're done, in the non-concurrent case */
  		heap_close(rel, NoLock);
! 		return;
  	}
  
  	/* save lockrelid and locktag for below, then close rel */
--- 618,624 ----
  	{
  		/* Close the heap and we're done, in the non-concurrent case */
  		heap_close(rel, NoLock);
! 		return indexRelationId;
  	}
  
  	/* save lockrelid and locktag for below, then close rel */
***************
*** 709,714 **** DefineIndex(RangeVar *heapRelation,
--- 906,913 ----
  	 * Last thing to do is release the session-level lock on the parent table.
  	 */
  	UnlockRelationIdForSession(&heaprelid, ShareUpdateExclusiveLock);
+ 
+ 	return indexRelationId;
  }
  
  
diff --git a/src/backend/commands/tableindex 3bc350a..a4430e2 100644
*** a/src/backend/commands/tablecmds.c
--- b/src/backend/commands/tablecmds.c
***************
*** 347,353 **** static bool ATColumnChangeRequiresRewrite(Node *expr, AttrNumber varattno);
  static void ATExecAlterColumnType(AlteredTableInfo *tab, Relation rel,
  					  AlterTableCmd *cmd, LOCKMODE lockmode);
  static void ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode);
! static void ATPostAlterTypeParse(char *cmd, List **wqueue, LOCKMODE lockmode);
  static void change_owner_recurse_to_sequences(Oid relationOid,
  								  Oid newOwnerId, LOCKMODE lockmode);
  static void ATExecClusterOn(Relation rel, const char *indexName, LOCKMODE lockmode);
--- 347,355 ----
  static void ATExecAlterColumnType(AlteredTableInfo *tab, Relation rel,
  					  AlterTableCmd *cmd, LOCKMODE lockmode);
  static void ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode);
! static void ATPostAlterTypeParse(Oid oldId, char *cmd,
! 					 List **wqueue, LOCKMODE lockmode, bool rewrite);
! static void TryReuseIndex(Oid oldId, IndexStmt *stmt);
  static void change_owner_recurse_to_sequences(Oid relationOid,
  								  Oid newOwnerId, LOCKMODE lockmode);
  static void ATExecClusterOn(Relation rel, const char *indexName, LOCKMODE lockmode);
***************
*** 5225,5261 **** ATExecAddIndex(AlteredTableInfo *tab, Relation rel,
  	bool		check_rights;
  	bool		skip_build;
  	bool		quiet;
  
  	Assert(IsA(stmt, IndexStmt));
  
  	/* suppress schema rights check when rebuilding existing index */
  	check_rights = !is_rebuild;
! 	/* skip index build if phase 3 will have to rewrite table anyway */
! 	skip_build = tab->rewrite;
  	/* suppress notices when rebuilding existing index */
  	quiet = is_rebuild;
  
  	/* The IndexStmt has already been through transformIndexStmt */
  
! 	DefineIndex(stmt->relation, /* relation */
! 				stmt->idxname,	/* index name */
! 				InvalidOid,		/* no predefined OID */
! 				stmt->accessMethod,		/* am name */
! 				stmt->tableSpace,
! 				stmt->indexParams,		/* parameters */
! 				(Expr *) stmt->whereClause,
! 				stmt->options,
! 				stmt->excludeOpNames,
! 				stmt->unique,
! 				stmt->primary,
! 				stmt->isconstraint,
! 				stmt->deferrable,
! 				stmt->initdeferred,
! 				true,			/* is_alter_table */
! 				check_rights,
! 				skip_build,
! 				quiet,
! 				false);
  }
  
  /*
--- 5227,5278 ----
  	bool		check_rights;
  	bool		skip_build;
  	bool		quiet;
+ 	Oid			new_index;
  
  	Assert(IsA(stmt, IndexStmt));
  
  	/* suppress schema rights check when rebuilding existing index */
  	check_rights = !is_rebuild;
! 	/* skip index build if phase 3 will do it or we're reusing an old one */
! 	skip_build = tab->rewrite || OidIsValid(stmt->oldNode);
  	/* suppress notices when rebuilding existing index */
  	quiet = is_rebuild;
  
  	/* The IndexStmt has already been through transformIndexStmt */
  
! 	new_index = DefineIndex(stmt->relation,		/* relation */
! 							stmt->idxname,		/* index name */
! 							InvalidOid, /* no predefined OID */
! 							stmt->oldNode,
! 							stmt->accessMethod, /* am name */
! 							stmt->tableSpace,
! 							stmt->indexParams,	/* parameters */
! 							(Expr *) stmt->whereClause,
! 							stmt->options,
! 							stmt->excludeOpNames,
! 							stmt->unique,
! 							stmt->primary,
! 							stmt->isconstraint,
! 							stmt->deferrable,
! 							stmt->initdeferred,
! 							true,		/* is_alter_table */
! 							check_rights,
! 							skip_build,
! 							quiet,
! 							false);
! 
! 	/*
! 	 * If TryReuseIndex() stashed a relfilenode for us, we used it for the new
! 	 * index instead of building from scratch.  The DROP of the old edition of
! 	 * this index will have scheduled the storage for deletion at commit, so
! 	 * cancel that pending deletion.
! 	 */
! 	if (OidIsValid(stmt->oldNode))
! 	{
! 		Relation	irel = index_open(new_index, NoLock);
! 		RelationPreserveStorage(irel->rd_node, true);
! 		index_close(irel, NoLock);
! 	}
  }
  
  /*
***************
*** 7375,7381 **** static void
  ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode)
  {
  	ObjectAddress obj;
! 	ListCell   *l;
  
  	/*
  	 * Re-parse the index and constraint definitions, and attach them to the
--- 7392,7399 ----
  ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode)
  {
  	ObjectAddress obj;
! 	ListCell   *def_item;
! 	ListCell   *oid_item;
  
  	/*
  	 * Re-parse the index and constraint definitions, and attach them to the
***************
*** 7385,7394 **** ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode)
  	 * that before dropping.  It's safe because the parser won't actually look
  	 * at the catalogs to detect the existing entry.
  	 */
! 	foreach(l, tab->changedIndexDefs)
! 		ATPostAlterTypeParse((char *) lfirst(l), wqueue, lockmode);
! 	foreach(l, tab->changedConstraintDefs)
! 		ATPostAlterTypeParse((char *) lfirst(l), wqueue, lockmode);
  
  	/*
  	 * Now we can drop the existing constraints and indexes --- constraints
--- 7403,7416 ----
  	 * that before dropping.  It's safe because the parser won't actually look
  	 * at the catalogs to detect the existing entry.
  	 */
! 	forboth(oid_item, tab->changedConstraintOids,
! 			def_item, tab->changedConstraintDefs)
! 		ATPostAlterTypeParse(lfirst_oid(oid_item), (char *) lfirst(def_item),
! 							 wqueue, lockmode, tab->rewrite);
! 	forboth(oid_item, tab->changedIndexOids,
! 			def_item, tab->changedIndexDefs)
! 		ATPostAlterTypeParse(lfirst_oid(oid_item), (char *) lfirst(def_item),
! 							 wqueue, lockmode, tab->rewrite);
  
  	/*
  	 * Now we can drop the existing constraints and indexes --- constraints
***************
*** 7398,7415 **** ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode)
  	 * should be okay to use DROP_RESTRICT here, since nothing else should be
  	 * depending on these objects.
  	 */
! 	foreach(l, tab->changedConstraintOids)
  	{
  		obj.classId = ConstraintRelationId;
! 		obj.objectId = lfirst_oid(l);
  		obj.objectSubId = 0;
  		performDeletion(&obj, DROP_RESTRICT);
  	}
  
! 	foreach(l, tab->changedIndexOids)
  	{
  		obj.classId = RelationRelationId;
! 		obj.objectId = lfirst_oid(l);
  		obj.objectSubId = 0;
  		performDeletion(&obj, DROP_RESTRICT);
  	}
--- 7420,7437 ----
  	 * should be okay to use DROP_RESTRICT here, since nothing else should be
  	 * depending on these objects.
  	 */
! 	foreach(oid_item, tab->changedConstraintOids)
  	{
  		obj.classId = ConstraintRelationId;
! 		obj.objectId = lfirst_oid(oid_item);
  		obj.objectSubId = 0;
  		performDeletion(&obj, DROP_RESTRICT);
  	}
  
! 	foreach(oid_item, tab->changedIndexOids)
  	{
  		obj.classId = RelationRelationId;
! 		obj.objectId = lfirst_oid(oid_item);
  		obj.objectSubId = 0;
  		performDeletion(&obj, DROP_RESTRICT);
  	}
***************
*** 7421,7427 **** ATPostAlterTypeCleanup(List **wqueue, AlteredTableInfo *tab, LOCKMODE lockmode)
  }
  
  static void
! ATPostAlterTypeParse(char *cmd, List **wqueue, LOCKMODE lockmode)
  {
  	List	   *raw_parsetree_list;
  	List	   *querytree_list;
--- 7443,7450 ----
  }
  
  static void
! ATPostAlterTypeParse(Oid oldId, char *cmd,
! 					 List **wqueue, LOCKMODE lockmode, bool rewrite)
  {
  	List	   *raw_parsetree_list;
  	List	   *querytree_list;
***************
*** 7468,7473 **** ATPostAlterTypeParse(char *cmd, List **wqueue, LOCKMODE lockmode)
--- 7491,7499 ----
  					IndexStmt  *stmt = (IndexStmt *) stm;
  					AlterTableCmd *newcmd;
  
+ 					if (!rewrite)
+ 						TryReuseIndex(oldId, stmt);
+ 
  					rel = relation_openrv(stmt->relation, lockmode);
  					tab = ATGetQueueEntry(wqueue, rel);
  					newcmd = makeNode(AlterTableCmd);
***************
*** 7492,7497 **** ATPostAlterTypeParse(char *cmd, List **wqueue, LOCKMODE lockmode)
--- 7518,7527 ----
  						switch (cmd->subtype)
  						{
  							case AT_AddIndex:
+ 								Assert(IsA(cmd->def, IndexStmt));
+ 								if (!rewrite)
+ 									TryReuseIndex(get_constraint_index(oldId),
+ 												  (IndexStmt *) cmd->def);
  								cmd->subtype = AT_ReAddIndex;
  								tab->subcmds[AT_PASS_OLD_INDEX] =
  									lappend(tab->subcmds[AT_PASS_OLD_INDEX], cmd);
***************
*** 7515,7520 **** ATPostAlterTypeParse(char *cmd, List **wqueue, LOCKMODE lockmode)
--- 7545,7570 ----
  	}
  }
  
+ /*
+  * Subroutine for ATPostAlterTypeParse().  Calls out to CheckIndexCompatible()
+  * for the real analysis, then mutates the IndexStmt based on that verdict.
+ */
+ static void
+ TryReuseIndex(Oid oldId, IndexStmt *stmt)
+ {
+ 
+ 	if (CheckIndexCompatible(oldId,
+ 							 stmt->relation,
+ 							 stmt->accessMethod,
+ 							 stmt->indexParams,
+ 							 stmt->excludeOpNames))
+ 	{
+ 		Relation irel = index_open(oldId, NoLock);
+ 		stmt->oldNode = irel->rd_node.relNode;
+ 		index_close(irel, NoLock);
+ 	}
+ }
+ 
  
  /*
   * ALTER TABLE OWNER
diff --git a/src/backend/nodes/copyfuncindex c9133dd..7a51456 100644
*** a/src/backend/nodes/copyfuncs.c
--- b/src/backend/nodes/copyfuncs.c
***************
*** 2807,2812 **** _copyIndexStmt(IndexStmt *from)
--- 2807,2813 ----
  	COPY_NODE_FIELD(whereClause);
  	COPY_NODE_FIELD(excludeOpNames);
  	COPY_SCALAR_FIELD(indexOid);
+ 	COPY_SCALAR_FIELD(oldNode);
  	COPY_SCALAR_FIELD(unique);
  	COPY_SCALAR_FIELD(primary);
  	COPY_SCALAR_FIELD(isconstraint);
diff --git a/src/backend/nodes/equalindex 3a0267c..4052a9a 100644
*** a/src/backend/nodes/equalfuncs.c
--- b/src/backend/nodes/equalfuncs.c
***************
*** 1245,1250 **** _equalIndexStmt(IndexStmt *a, IndexStmt *b)
--- 1245,1251 ----
  	COMPARE_NODE_FIELD(whereClause);
  	COMPARE_NODE_FIELD(excludeOpNames);
  	COMPARE_SCALAR_FIELD(indexOid);
+ 	COMPARE_SCALAR_FIELD(oldNode);
  	COMPARE_SCALAR_FIELD(unique);
  	COMPARE_SCALAR_FIELD(primary);
  	COMPARE_SCALAR_FIELD(isconstraint);
diff --git a/src/backend/nodes/outfunindex 681f5f8..b5be09a 100644
*** a/src/backend/nodes/outfuncs.c
--- b/src/backend/nodes/outfuncs.c
***************
*** 1976,1981 **** _outIndexStmt(StringInfo str, IndexStmt *node)
--- 1976,1982 ----
  	WRITE_NODE_FIELD(whereClause);
  	WRITE_NODE_FIELD(excludeOpNames);
  	WRITE_OID_FIELD(indexOid);
+ 	WRITE_OID_FIELD(oldNode);
  	WRITE_BOOL_FIELD(unique);
  	WRITE_BOOL_FIELD(primary);
  	WRITE_BOOL_FIELD(isconstraint);
diff --git a/src/backend/tcop/utiliindex 224e1f3..d13a56e 100644
*** a/src/backend/tcop/utility.c
--- b/src/backend/tcop/utility.c
***************
*** 945,950 **** standard_ProcessUtility(Node *parsetree,
--- 945,951 ----
  				DefineIndex(stmt->relation,		/* relation */
  							stmt->idxname,		/* index name */
  							InvalidOid, /* no predefined OID */
+ 							InvalidOid, /* no previous storage */
  							stmt->accessMethod, /* am name */
  							stmt->tableSpace,
  							stmt->indexParams,	/* parameters */
diff --git a/src/backend/utils/caindex 0b9d77a..809222b 100644
*** a/src/backend/utils/cache/relcache.c
--- b/src/backend/utils/cache/relcache.c
***************
*** 2395,2400 **** RelationBuildLocalRelation(const char *relname,
--- 2395,2401 ----
  						   Oid relnamespace,
  						   TupleDesc tupDesc,
  						   Oid relid,
+ 						   Oid relfilenode,
  						   Oid reltablespace,
  						   bool shared_relation,
  						   bool mapped_relation,
***************
*** 2529,2538 **** RelationBuildLocalRelation(const char *relname,
  
  	/*
  	 * Insert relation physical and logical identifiers (OIDs) into the right
! 	 * places.	Note that the physical ID (relfilenode) is initially the same
! 	 * as the logical ID (OID); except that for a mapped relation, we set
! 	 * relfilenode to zero and rely on RelationInitPhysicalAddr to consult the
! 	 * map.
  	 */
  	rel->rd_rel->relisshared = shared_relation;
  
--- 2530,2537 ----
  
  	/*
  	 * Insert relation physical and logical identifiers (OIDs) into the right
! 	 * places.  For a mapped relation, we set relfilenode to zero and rely on
! 	 * RelationInitPhysicalAddr to consult the map.
  	 */
  	rel->rd_rel->relisshared = shared_relation;
  
***************
*** 2547,2556 **** RelationBuildLocalRelation(const char *relname,
  	{
  		rel->rd_rel->relfilenode = InvalidOid;
  		/* Add it to the active mapping information */
! 		RelationMapUpdateMap(relid, relid, shared_relation, true);
  	}
  	else
! 		rel->rd_rel->relfilenode = relid;
  
  	RelationInitLockInfo(rel);	/* see lmgr.c */
  
--- 2546,2555 ----
  	{
  		rel->rd_rel->relfilenode = InvalidOid;
  		/* Add it to the active mapping information */
! 		RelationMapUpdateMap(relid, relfilenode, shared_relation, true);
  	}
  	else
! 		rel->rd_rel->relfilenode = relfilenode;
  
  	RelationInitLockInfo(rel);	/* see lmgr.c */
  
diff --git a/src/backend/utils/cache/relmindex a19ee28..b04dc9e 100644
*** a/src/backend/utils/cache/relmapper.c
--- b/src/backend/utils/cache/relmapper.c
***************
*** 792,798 **** write_relmap_file(bool shared, RelMapFile *newmap,
  			rnode.spcNode = tsid;
  			rnode.dbNode = dbid;
  			rnode.relNode = newmap->mappings[i].mapfilenode;
! 			RelationPreserveStorage(rnode);
  		}
  	}
  
--- 792,798 ----
  			rnode.spcNode = tsid;
  			rnode.dbNode = dbid;
  			rnode.relNode = newmap->mappings[i].mapfilenode;
! 			RelationPreserveStorage(rnode, false);
  		}
  	}
  
diff --git a/src/include/catalog/heap.h b/index 0b7b190..aee2d88 100644
*** a/src/include/catalog/heap.h
--- b/src/include/catalog/heap.h
***************
*** 39,44 **** extern Relation heap_create(const char *relname,
--- 39,45 ----
  			Oid relnamespace,
  			Oid reltablespace,
  			Oid relid,
+ 			Oid relfilenode,
  			TupleDesc tupDesc,
  			char relkind,
  			char relpersistence,
diff --git a/src/include/catalog/index 071db7f..8b78b05 100644
*** a/src/include/catalog/index.h
--- b/src/include/catalog/index.h
***************
*** 35,40 **** extern void index_check_primary_key(Relation heapRel,
--- 35,41 ----
  extern Oid index_create(Relation heapRelation,
  			 const char *indexRelationName,
  			 Oid indexRelationId,
+ 			 Oid relFileNode,
  			 IndexInfo *indexInfo,
  			 List *indexColNames,
  			 Oid accessMethodObjectId,
diff --git a/src/include/catalog/sindex 8dee8cf..6907d83 100644
*** a/src/include/catalog/storage.h
--- b/src/include/catalog/storage.h
***************
*** 22,28 ****
  
  extern void RelationCreateStorage(RelFileNode rnode, char relpersistence);
  extern void RelationDropStorage(Relation rel);
! extern void RelationPreserveStorage(RelFileNode rnode);
  extern void RelationTruncate(Relation rel, BlockNumber nblocks);
  
  /*
--- 22,28 ----
  
  extern void RelationCreateStorage(RelFileNode rnode, char relpersistence);
  extern void RelationDropStorage(Relation rel);
! extern void RelationPreserveStorage(RelFileNode rnode, bool atCommit);
  extern void RelationTruncate(Relation rel, BlockNumber nblocks);
  
  /*
diff --git a/src/include/commands/deindex bbc024f..81c515e 100644
*** a/src/include/commands/defrem.h
--- b/src/include/commands/defrem.h
***************
*** 18,26 ****
  
  
  /* commands/indexcmds.c */
! extern void DefineIndex(RangeVar *heapRelation,
  			char *indexRelationName,
  			Oid indexRelationId,
  			char *accessMethodName,
  			char *tableSpaceName,
  			List *attributeList,
--- 18,27 ----
  
  
  /* commands/indexcmds.c */
! extern Oid DefineIndex(RangeVar *heapRelation,
  			char *indexRelationName,
  			Oid indexRelationId,
+ 			Oid relFileNode,
  			char *accessMethodName,
  			char *tableSpaceName,
  			List *attributeList,
***************
*** 49,54 **** extern char *ChooseIndexName(const char *tabname, Oid namespaceId,
--- 50,60 ----
  				List *colnames, List *exclusionOpNames,
  				bool primary, bool isconstraint);
  extern List *ChooseIndexColumnNames(List *indexElems);
+ extern bool CheckIndexCompatible(Oid oldId,
+ 					 RangeVar *heapRelation,
+ 					 char *accessMethodName,
+ 					 List *attributeList,
+ 					 List *exclusionOpNames);
  extern Oid	GetDefaultOpClass(Oid type_id, Oid am_id);
  
  /* commands/functioncmds.c */
diff --git a/src/include/nodes/parseindex 00c1269..92e40d3 100644
*** a/src/include/nodes/parsenodes.h
--- b/src/include/nodes/parsenodes.h
***************
*** 2062,2067 **** typedef struct IndexStmt
--- 2062,2068 ----
  	Node	   *whereClause;	/* qualification (partial-index predicate) */
  	List	   *excludeOpNames; /* exclusion operator names, or NIL if none */
  	Oid			indexOid;		/* OID of an existing index, if any */
+ 	Oid			oldNode;		/* relfilenode of my former self */
  	bool		unique;			/* is index unique? */
  	bool		primary;		/* is index on primary key? */
  	bool		isconstraint;	/* is it from a CONSTRAINT clause? */
diff --git a/src/include/utils/relcacindex 1f4def5..ef1692c 100644
*** a/src/include/utils/relcache.h
--- b/src/include/utils/relcache.h
***************
*** 67,72 **** extern Relation RelationBuildLocalRelation(const char *relname,
--- 67,73 ----
  						   Oid relnamespace,
  						   TupleDesc tupDesc,
  						   Oid relid,
+ 						   Oid relfilenode,
  						   Oid reltablespace,
  						   bool shared_relation,
  						   bool mapped_relation,
diff --git a/src/test/regress/expecindex 8e75b14..44df2ea 100644
diff --git a/src/test/regress/sql/big_alter_table.sqindex b568302..6ae9007 100644
